@@ -40,6 +40,10 @@ def _lookup_schema(name):
     return eval(expr, mod.__dict__)
 
 
+def _prefix(name):
+    return (u'X%s:' % name.upper()).encode('utf-8')
+
+
 class Record(object):
 
     def __init__(self, schema):
@@ -226,6 +230,16 @@ class Database(object):
 
     value_count = property(_get_value_count, _set_value_count)
 
+    def check_prefix(self, name, boolean=False):
+        prefixes = set(self.relevance_prefixes.keys())
+        prefixes.union(self.boolean_prefixes.keys())
+        if name not in prefixes:
+            upped = _prefix(name)
+            if boolean:
+                self.add_boolean_prefix(name, upped)
+            else:
+                self.add_prefix(name, upped)
+
     def add_prefix(self, key, value):
         """Add a prefix mapping to the database.
         """
@@ -329,9 +343,70 @@ class Database(object):
                                   (schema.__class__.__name__,
                                    list((c.name, c.value)
                                         for c in invalid)))
-        doc = schema.to_document(self)
-        data = dumps((_schema_name(schema_type), schema.flatten()))
-        doc.set_data(data)
+
+        doc = self.doc_from_dict(schema.memo.dict)
+        return doc
+
+    def doc_from_dict(self, data):
+        """
+        """
+        doc = xapian.Document()
+
+        for term_tup in data.get('terms', ()):
+            term = None
+            typ = None
+            wdfinc = 1
+            if len(term_tup) > 2:
+                term, typ, wdfinc = term_tup
+            else:
+                term, typ = term_tup
+            if ':' in term:
+                prefix, _, value = term.partition(':')
+                self.check_prefix(prefix, typ)
+                term = 'X%s:%s' % (prefix.upper(), value)
+            doc.add_term(term, wdfinc)
+                
+        for post in data.get('posts', ()):
+            pass
+
+        for text_dict in data.get('texts', ()):
+            lang = text_dict.get('lang')
+            tg = xapian.TermGenerator()
+            tg.set_database(self.backend)
+            tg.set_document(doc)
+            spelling = text_dict.get('spell', True)
+            if spelling:
+                tg.set_flags(xapian.TermGenerator.FLAG_SPELLING)
+            if lang in snowball.stoppers:
+                tg.set_stemmer(xapian.Stem(lang))
+                tg.set_stopper(snowball.stoppers[lang])
+            if text_dict.get('post', True):
+                index_text = tg.index_text
+            else:
+                index_text = tg.index_text_without_positions
+            text = text_dict.get('text')
+            if text:
+                prefix = text_dict.get('prefix')
+                if prefix:
+                    self.check_prefix(prefix)
+                    prefix = 'X%s:' % prefix.upper()
+                    index_text(text, 1, prefix)
+                else:
+                    index_text(text)
+
+        for val_tuple in data.get('values', ()):
+            name, value, typ = val_tuple
+            if name in self.values:
+                valno = self.values[name]
+            else:
+                valno = self.add_value(name, typ)
+            if typ in ('integer',):
+                value = xapian.sortable_serialise(value)
+            doc.add_value(valno, value)
+
+        data = data.get('data')
+        if data:
+            doc.set_data(data)
         return doc
 
     def reopen(self):
