@@ -1,10 +1,10 @@
-import xapian
 import unicodedata
 import translitcodec
 
 from flatland import schema
+from json import dumps
 
-from . import snowball
+from . memo import Memo
 from . tools import geoprint
 
 from .exc import (
@@ -15,13 +15,20 @@ from .exc import (
 MAX_TERM_LEN = 240 # compile time xapian limit
 
 
-def _prefix(name):
-    return (u'X%s:' % name.upper()).encode('utf-8')
+def _schema_name(schema):
+    return "%s.%s" % (schema.__module__, schema.__class__.__name__)
+
+
+def _prefix(prefix, value=None):
+    if prefix is None:
+        return
+    return ("%s:%s" % (prefix, value)) if value else prefix
 
 def _normalize(value):
     return unicodedata.normalize(
         'NFKC', unicode(value).strip().lower()).encode('utf-8')
 
+
 _use_schema = object() # marker says use schema term generator
 
 
@@ -38,24 +45,6 @@ class Schema(schema.Form):
     own.
     """
 
-    database = None
-    """The xodb database being indexed into.
-
-    Initialized by to_document()
-    """
-
-    document = None
-    """The xapian document the schema will index into.
-
-    Initialized by to_document()
-    """
-
-    term_gen = None
-    """The xapian term generator for indexing text.
-
-    Initialized by to_document()
-    """
-
     facet_prefix = 'facet'
     """The default prefix that will be used for storing and searching
     for facets.
@@ -76,24 +65,24 @@ class Schema(schema.Form):
             elif hasattr(obj, name):
                 self[name] = getattr(obj, name)
 
-    def to_document(self, database):
+    @property
+    def memo(self):
         """
         Walk the children, indexing each one with a handler into a
-        Xapian document.
-
-        The document is not added to the database, it is the
-        responsibility of the caller to do that.
+        memo object.
         """
-        language = (self.language or
-                    self.root.get('language').value or
-                    self.root.get('language').default)
-
-        self.database = database
-        self.document = xapian.Document()
-        self.term_gen = self._init_term_gen(language)
+        lang = None
+        try:
+            root_lang = self.root.get('language')
+            lang = root_lang.value or root_lang.default
+        except KeyError:
+            lang = self.language
 
+        self._memo = Memo()
+        self._memo.set_lang(lang)
         self._handle_children(self)
-        return self.document
+        self._memo.data = dumps((_schema_name(self), self.flatten()))
+        return self._memo
 
     def _handle_children(self, parent):
         for el in parent.all_children:
@@ -109,7 +98,7 @@ class Schema(schema.Form):
                 elif isinstance(el, Date):
                     handler = self._handle_date
                 elif isinstance(el, DateTime):
-                    handler = self._handle_date
+                    handler = self._handle_datetime
                 elif isinstance(el, Text):
                     handler = self._handle_text
                 elif isinstance(el, NumericRange):
@@ -129,35 +118,33 @@ class Schema(schema.Form):
 
                 if value is not None:
                     if el.facet:
-                        self._add_term(el.name, self.facet_prefix, True)
+                        self._memo.add_term(
+                            _prefix(self.facet_prefix, el.name),
+                            True)
         return True
 
-    def _handle_scalar(self, term, value, element,
-                       type=None, serialize=False):
+    def _handle_scalar(self, term, value, element, type=None):
         name = element.flattened_name()
+        memo = self._memo
         if element.prefix:
-            self._add_term(term, name, element.boolean)
+            memo.add_term(_prefix(name, term), element.boolean)
         else:
-            self._add_term(term)
+            memo.add_term(term, element.boolean)
         if element.sortable:
-            if serialize:
-                self._add_value(
-                    xapian.sortable_serialise(value),
-                    name,
-                    type)
-            else:
-                self._add_value(value, name, type)
+            memo.add_value(name, value, type)
         return value
 
     def _handle_string(self, element):
+        term = element.u
         value = element.value
         if value:
-            return self._handle_scalar(value, value, element, 'string')
+            return self._handle_scalar(term, value, element, 'string')
 
     def _handle_integer(self, element):
+        term = element.u
         value = element.value
         if value:
-            return self._handle_scalar(value, value, element, 'integer', True)
+            return self._handle_scalar(term, value, element, 'integer')
 
     def _handle_float(self, element):
         # TODO:mp floats are currently storage only
@@ -166,7 +153,7 @@ class Schema(schema.Form):
     def _handle_boolean(self, element):
         value = 'true' if element.value else 'false'
         if value:
-            return self._handle_scalar(value, value, element, 'integer', True)
+            return self._handle_scalar(value, value, element, 'integer')
 
     def _handle_date(self, element):
         if element.value:
@@ -174,39 +161,46 @@ class Schema(schema.Form):
             value = element.value.strftime(element.value_format)
             return self._handle_scalar(term, value, element, 'date')
 
+    def _handle_datetime(self, element):
+        if element.value:
+            term = element.value.strftime(element.term_format)
+            value = element.value.strftime(element.value_format)
+            return self._handle_scalar(term, value, element, 'datetime')
+
     def _handle_text(self, element):
         value = element.value
         if value is None:
             return
         name = element.flattened_name()
-        if element.language is _use_schema:
-            tg = self.term_gen
-        else:
-            tg = self._init_term_gen(element.language)
+        memo = self._memo
 
-        if element.positions:
-            index_text = tg.index_text
+        if element.language is _use_schema:
+            lang = memo.lang
         else:
-            index_text = tg.index_text_without_positions
+            lang = element.language
 
         if element.translit:
             value = value.encode(element.translit)
 
         if element.string:
             if element.string_prefix:
-                self._add_term(value, element.string_prefix, element.boolean)
+                memo.add_term(_prefix(element.string_prefix, value),
+                               element.boolean)
             else:
-                self._add_term(value)
+                memo.add_term(value)
 
         if element.sortable:
-            self._add_value(value, name, 'string')
+            memo.add_value(name, value, 'string')
 
         value = _normalize(value)
+        prefix = None
         if element.prefix:
-            self._check_prefix(name, element.boolean)
-            index_text(value, 1, _prefix(name))
-        else:
-            index_text(value)
+            prefix = _prefix(name)
+        memo.add_text(value, prefix, lang,
+                      element.positions,
+                      element.stem,
+                      element.stop,
+                      element.spelling)
         return value
 
     def _handle_numericrange(self, element):
@@ -226,55 +220,20 @@ class Schema(schema.Form):
 
         for i in xrange(minv, maxv, step):
             val = "%s_%s" % (i, i + step)
-            self._add_term(val, element.name, True)
+            self._memo.add_term(_prefix(element.name, val), True)
         return True
 
     def _handle_location(self, element):
+        memo = self._memo
         h = 'loc_' + element.hash(element.radians)
-        self._add_term(h, boolean=True)
+        if element.boolean:
+            memo.add_term(h, boolean=True)
+        else:
+            memo.add_term(h)
+        if element.sortable:
+            memo.add_value(element.name, h, 'location')
         return True
 
-    def _check_prefix(self, name, boolean=False):
-        database = self.database
-        prefixes = database.relevance_prefixes.copy()
-        prefixes.update(database.boolean_prefixes.copy())
-        if name not in prefixes:
-            upped = _prefix(name)
-            if boolean:
-                database.add_boolean_prefix(name, upped)
-            else:
-                database.add_prefix(name, upped)
-
-    def _init_term_gen(self, language):
-        tg = xapian.TermGenerator()
-        tg.set_database(self.database.backend)
-        tg.set_document(self.document)
-        if self.database.spelling:
-            tg.set_flags(xapian.TermGenerator.FLAG_SPELLING)
-        if language in snowball.stoppers:
-            tg.set_stemmer(xapian.Stem(language))
-            tg.set_stopper(snowball.stoppers[language])
-        return tg
-
-    def _add_term(self, value, prefix=None, boolean=False):
-        value = _normalize(value)
-        if len(value) > MAX_TERM_LEN:
-            raise InvalidTermError('The term %s is too long.' % value)
-        if prefix:
-            self._check_prefix(prefix, boolean)
-            term = _prefix(prefix) + value.lower()
-        else:
-            term = value.lower()
-        self.document.add_term(term)
-
-    def _add_value(self, value, name, type):
-        database = self.database
-        if name in database.values:
-            valno = database.values[name]
-        else:
-            valno = database.add_value(name, type)
-        self.document.add_value(valno, value)
-
 
 class _BaseElement(object):
 
@@ -417,6 +376,18 @@ class Text(schema.String, _BaseElement):
     of a larger database.
     """
 
+    stem = True
+    """If the language is set, stem the text.
+    """
+
+    stop = True
+    """If the language is set, remove stopwords from text.
+    """
+
+    spelling = True
+    """True if the text should be added to the db as spelling correction targets.
+    """
+
     string_prefix = None
     """If not None, defines the prefix that should be used for string
     indexing the text's value.
@@ -493,7 +464,7 @@ class Location(schema.Compound, _BaseElement):
     def compose(self):
         numbers = (self['lat'].value, self['lon'].value)
         display = ' by '.join(str(n) for n in numbers)
-        return display, numbers
+        return numbers, display
 
     def explode(self, value):
         self['lat'] = value[0]
